// Copyright (c) 2023 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#pragma once
#include "dawn.hpp"
#include "display/RenderTarget.hpp"
#include "display/RenderPipeline.hpp"
#include "display/shader/ShaderManager.hpp"

namespace Dawn {
  class Game;

  class IRenderHost {
    public:
      RenderPipeline renderPipeline;
      ShaderManager shaderManager;

      /**
       * Creates a render host.
       */
      IRenderHost();

      /**
       * Initializes the render host, called by the game during the initial 
       * set up of the engine.
       * 
       * @param game Game that requested the render host to initialize.
       */
      virtual void init(const std::shared_ptr<Game> game) = 0;

      /**
       * Performs an update/tick of the render host. This would be the game 
       * asking the RenderHost to do the rendering.
       */
      virtual void update(const std::shared_ptr<Game> game) = 0;

      /**
       * Overridable request from the game that asks if the RenderHost has any
       * reason that it should need to close. For most libraries this would be
       * whether or not the window was closed.
       * 
       * @return True if the render host requests the game to gracefully exit.
       */
      virtual bool_t isCloseRequested() = 0;

      /**
       * Returns the back buffer render target. This is the render target that
       * is used to render to the screen.
       * 
       * @return The back buffer render target.
       */
      virtual std::shared_ptr<RenderTarget> getBackBufferRenderTarget() = 0;

      /**
       * Destroys the render host.
       */
      virtual ~IRenderHost();
  };
}